#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import tempfile
import time
from pathlib import Path

import pytest

from e2e_tests.base import AVAILABLE_CONTAINERS, BaseE2ETest, fake
from e2e_tests.utils import tmp_context, with_check_if_file_appears
from neptune.new.metadata_containers import MetadataContainer


class TestArtifacts(BaseE2ETest):
    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_local_creation(self, container: MetadataContainer):
        first, second = self.gen_key(), self.gen_key()
        filename = fake.unique.file_name()

        with tmp_context() as tmp:
            with open(filename, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            container[first].track_files(".")
            container[second].track_files(f"file://{tmp}")

            container.sync()

        assert container[first].fetch_hash() == container[second].fetch_hash()
        assert container[first].fetch_files_list() == container[second].fetch_files_list()

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_assignment(self, container: MetadataContainer):
        first, second = self.gen_key(), self.gen_key()
        filename = fake.unique.file_name()

        with tmp_context():
            with open(filename, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            container[first].track_files(filename)
            container.wait()
            container[second] = container[first].fetch()
            container.sync()

        assert container[first].fetch_hash() == container[second].fetch_hash()
        assert container[first].fetch_files_list() == container[second].fetch_files_list()

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_local_download(self, container: MetadataContainer):
        first, second = self.gen_key(), self.gen_key()
        filename, filepath = fake.unique.file_name(), fake.unique.file_path(depth=3).lstrip("/")

        with tmp_context() as tmp:
            with open(filename, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            os.makedirs(Path(filepath).parent, exist_ok=True)
            with open(filepath, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            # Relative path
            container[first].track_files(filename)
            # Absolute path
            container[second].track_files(tmp)

            container.sync()

            with tmp_context():
                with with_check_if_file_appears(f"artifacts/{filename}"):
                    container[first].download("artifacts/")

                with with_check_if_file_appears(filepath):
                    container[second].download()

    @pytest.mark.s3
    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_s3_creation(self, container: MetadataContainer, bucket, environment):
        first, second, prefix = (
            self.gen_key(),
            self.gen_key(),
            f"{environment.project}/{self.gen_key()}/{type(container).__name__}",
        )
        filename = fake.unique.file_name()

        bucket_name, s3_client = bucket

        with tmp_context():
            with open(filename, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            s3_client.meta.client.upload_file(filename, bucket_name, f"{prefix}/{filename}")

        container[first].track_files(f"s3://{bucket_name}/{prefix}/{filename}")
        container[second].track_files(f"s3://{bucket_name}/{prefix}")

        container.sync()

        assert container[first].fetch_hash() == container[second].fetch_hash()
        assert container[first].fetch_files_list() == container[second].fetch_files_list()

    @pytest.mark.s3
    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_s3_download(self, container: MetadataContainer, bucket, environment):
        first = self.gen_key()
        prefix = f"{environment.project}/{self.gen_key()}/{type(container).__name__}"
        filename, filepath = fake.unique.file_name(), fake.unique.file_path(depth=3).lstrip("/")

        bucket_name, s3_client = bucket

        with tmp_context():
            with open(filename, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            os.makedirs(Path(filepath).parent, exist_ok=True)
            with open(filepath, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            s3_client.meta.client.upload_file(filename, bucket_name, f"{prefix}/{filename}")
            s3_client.meta.client.upload_file(filepath, bucket_name, f"{prefix}/{filepath}")

        container[first].track_files(f"s3://{bucket_name}/{prefix}")

        container.sync()

        with tempfile.TemporaryDirectory() as tmp:
            with with_check_if_file_appears(f"{tmp}/{filename}"):
                container[first].download(tmp)

        with tmp_context():
            with with_check_if_file_appears(filename):
                container[first].download()

    @pytest.mark.s3
    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_s3_existing(self, container: MetadataContainer, bucket, environment):
        first, second, prefix = (
            self.gen_key(),
            self.gen_key(),
            f"{environment.project}/{self.gen_key()}/{type(container).__name__}",
        )
        filename, filepath = fake.file_name(), fake.file_path(depth=3).lstrip("/")

        bucket_name, s3_client = bucket

        with tmp_context():
            with open(filename, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            os.makedirs(Path(filepath).parent, exist_ok=True)
            with open(filepath, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            s3_client.meta.client.upload_file(filename, bucket_name, f"{prefix}/{filename}")
            s3_client.meta.client.upload_file(filepath, bucket_name, f"{prefix}/{filepath}")

        # Track all files - "a" and "b" to first artifact
        container[first].track_files(f"s3://{bucket_name}/{prefix}/")

        # Track only the "a" file to second artifact
        container[second].track_files(f"s3://{bucket_name}/{prefix}/{filename}")
        container.sync()

        # Add "b" file to existing second artifact
        # so it should be now identical as first
        container[second].track_files(
            f"s3://{bucket_name}/{prefix}/{filepath}",
            destination=str(Path(filepath).parent),
        )
        container.sync()

        assert container[first].fetch_hash() == container[second].fetch_hash()
        assert container[first].fetch_files_list() == container[second].fetch_files_list()

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_local_existing(self, container: MetadataContainer):
        first, second = self.gen_key(), self.gen_key()
        filename, filepath = fake.file_name(), fake.file_path(depth=3).lstrip("/")

        with tmp_context() as tmp:
            with open(filename, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            os.makedirs(Path(filepath).parent, exist_ok=True)
            with open(filepath, "w", encoding="utf-8") as handler:
                handler.write(fake.paragraph(nb_sentences=5))

            # Track all files - "a" and "b" to first artifact
            container[first].track_files(".")

            # Track only the "a" file to second artifact
            container[second].track_files(f"file://{tmp}/{filename}")
            container.sync()

            # Add "b" file to existing second artifact
            # so it should be now identical as first
            container[second].track_files(filepath, destination=str(Path(filepath).parent))
            container.sync()

        assert container[first].fetch_hash() == container[second].fetch_hash()
        assert container[first].fetch_files_list() == container[second].fetch_files_list()

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_hash_cache(self, container: MetadataContainer):
        key = self.gen_key()
        filename = fake.file_name()

        with tmp_context():
            # create 2GB file
            with open(filename, "wb") as handler:
                handler.write(b"\0" * 2 * 2**30)

            # track it
            start = time.time()
            container[key].track_files(".", wait=True)
            initial_duration = time.time() - start

            # and track it again
            start = time.time()
            container[key].track_files(".", wait=True)
            retry_duration = time.time() - start

            assert (
                retry_duration * 2 < initial_duration
            ), "Tracking again should be significantly faster"

            # append additional byte to file
            with open(filename, "ab") as handler:
                handler.write(b"\0")

            # and track updated file
            start = time.time()
            container[key].track_files(".", wait=True)
            updated_duration = time.time() - start

            assert (
                retry_duration * 2 < updated_duration
            ), "Tracking updated file should take more time - no cache"



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
import uuid
from unittest.mock import Mock

from neptune.new.attributes import Integer, String
from neptune.new.exceptions import FetchAttributeNotFoundException
from neptune.new.internal import operation
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.backends.utils import (
    ExecuteOperationsBatchingManager,
    build_operation_url,
)
from neptune.new.internal.container_type import ContainerType


class TestNeptuneBackendMock(unittest.TestCase):
    def test_building_operation_url(self):
        urls = {
            build_operation_url(
                "https://app.stage.neptune.ai", "api/leaderboard/v1/attributes/download"
            ),
            build_operation_url(
                "https://app.stage.neptune.ai",
                "/api/leaderboard/v1/attributes/download",
            ),
            build_operation_url(
                "https://app.stage.neptune.ai/",
                "api/leaderboard/v1/attributes/download",
            ),
            build_operation_url(
                "https://app.stage.neptune.ai/",
                "/api/leaderboard/v1/attributes/download",
            ),
            build_operation_url("app.stage.neptune.ai", "api/leaderboard/v1/attributes/download"),
            build_operation_url("app.stage.neptune.ai", "/api/leaderboard/v1/attributes/download"),
            build_operation_url("app.stage.neptune.ai/", "api/leaderboard/v1/attributes/download"),
            build_operation_url("app.stage.neptune.ai/", "/api/leaderboard/v1/attributes/download"),
        }
        self.assertEqual(
            {"https://app.stage.neptune.ai/api/leaderboard/v1/attributes/download"},
            urls,
        )


class TestExecuteOperationsBatchingManager(unittest.TestCase):
    def test_cut_batch_on_copy(self):
        backend = Mock(spec=NeptuneBackend)
        manager = ExecuteOperationsBatchingManager(backend)

        operations = [
            operation.AssignInt(["a"], 12),
            operation.AssignString(["b/c"], "test"),
            operation.CopyAttribute(["a"], str(uuid.uuid4()), ContainerType.RUN, ["b"], Integer),
            operation.AssignFloat(["q/d"], 44.12),
            operation.AssignInt(["pp"], 12),
            operation.CopyAttribute(["q/d"], str(uuid.uuid4()), ContainerType.RUN, ["b"], String),
        ]

        batch = manager.get_batch(operations)
        self.assertEqual(operations[0:2], batch.operations)
        self.assertEqual([], batch.errors)
        self.assertEqual(0, batch.dropped_operations_count)

    def test_get_nonempty_batch_with_copy_first(self):
        backend = Mock(spec=NeptuneBackend)
        manager = ExecuteOperationsBatchingManager(backend)

        operations = [
            operation.CopyAttribute(["a"], str(uuid.uuid4()), ContainerType.RUN, ["b"], Integer),
            operation.AssignFloat(["q/d"], 44.12),
            operation.AssignInt(["pp"], 12),
            operation.CopyAttribute(["q/d"], str(uuid.uuid4()), ContainerType.RUN, ["b"], String),
        ]

        batch = manager.get_batch(operations)
        expected_batch = [
            operation.AssignInt(operations[0].path, backend.get_int_attribute.return_value.value)
        ] + operations[1:3]
        self.assertEqual(expected_batch, batch.operations)
        self.assertEqual([], batch.errors)
        self.assertEqual(0, batch.dropped_operations_count)

    def test_no_copies_is_ok(self):
        backend = Mock(spec=NeptuneBackend)
        manager = ExecuteOperationsBatchingManager(backend)

        operations = [
            operation.AssignInt(["a"], 12),
            operation.AssignString(["b/c"], "test"),
            operation.AssignFloat(["q/d"], 44.12),
            operation.AssignInt(["pp"], 12),
        ]

        batch = manager.get_batch(operations)
        self.assertEqual(operations, batch.operations)
        self.assertEqual([], batch.errors)
        self.assertEqual(0, batch.dropped_operations_count)

    def test_no_ops_is_ok(self):
        backend = Mock(spec=NeptuneBackend)
        manager = ExecuteOperationsBatchingManager(backend)

        batch = manager.get_batch([])
        self.assertEqual([], batch.operations)
        self.assertEqual([], batch.errors)
        self.assertEqual(0, batch.dropped_operations_count)

    def test_subsequent_copies_is_ok(self):
        backend = Mock(spec=NeptuneBackend)
        manager = ExecuteOperationsBatchingManager(backend)

        operations = [
            operation.CopyAttribute(["a"], str(uuid.uuid4()), ContainerType.RUN, ["b"], Integer),
            operation.CopyAttribute(["q/d"], str(uuid.uuid4()), ContainerType.RUN, ["b"], String),
            operation.CopyAttribute(["pp"], str(uuid.uuid4()), ContainerType.RUN, ["b"], Integer),
        ]

        batch = manager.get_batch(operations)
        expected_batch = [
            operation.AssignInt(operations[0].path, backend.get_int_attribute.return_value.value)
        ]
        self.assertEqual(expected_batch, batch.operations)
        self.assertEqual([], batch.errors)
        self.assertEqual(0, batch.dropped_operations_count)

    def test_handle_failed_copy(self):
        backend = Mock(spec=NeptuneBackend)
        backend.get_int_attribute.side_effect = FetchAttributeNotFoundException("b")
        manager = ExecuteOperationsBatchingManager(backend)

        operations = [
            operation.CopyAttribute(["q/d"], str(uuid.uuid4()), ContainerType.RUN, ["b"], Integer),
            operation.AssignInt(["a"], 12),
            operation.AssignString(["b/c"], "test"),
            operation.AssignInt(["pp"], 12),
        ]

        batch = manager.get_batch(operations)
        # skipped erroneous CopyAttribute
        self.assertEqual(operations[1:], batch.operations)
        self.assertEqual([backend.get_int_attribute.side_effect], batch.errors)
        self.assertEqual(1, batch.dropped_operations_count)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Remember to set environment values:
        # self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].pop('tag2_to_remove')
        # self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].pop('tag4_remove_non_existing')
* NEPTUNE_API_TOKEN
* NEPTUNE_PROJECT
"""
import sys
from datetime import datetime

from PIL import Image

import neptune.new as neptune
from alpha_integration_dev.common_client_code import ClientFeatures
from neptune.new.attributes.constants import (
    ARTIFACT_ATTRIBUTE_SPACE,
    LOG_ATTRIBUTE_SPACE,
    PROPERTIES_ATTRIBUTE_SPACE,
    SOURCE_CODE_FILES_ATTRIBUTE_PATH,
    SYSTEM_TAGS_ATTRIBUTE_PATH,
)
from neptune.new.exceptions import MissingFieldException
from neptune.new.types import File


class NewClientFeatures(ClientFeatures):
    def __init__(self):
        super().__init__()
        self.exp = neptune.init(
            source_files="alpha_integration_dev/*.py",
        )

        # download sources
        self.exp.sync()
        with self.with_check_if_file_appears("files.zip"):
            self.exp[SOURCE_CODE_FILES_ATTRIBUTE_PATH].download()

    def modify_tags(self):
        self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].add("tag1")
        self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].add(["tag2_to_remove", "tag3"])
        self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].remove("tag2_to_remove")
        self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].remove("tag4_remove_non_existing")
        # del self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH]  # TODO: NPT-9222

        self.exp.sync()
        assert self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].fetch() == {"tag1", "tag3"}

    def modify_properties(self):
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop"] = "some text"
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop_number"] = 42
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]["nested/prop"] = 42
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop_to_del"] = 42
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop_list"] = [1, 2, 3]
        with open(self.text_file_path, mode="r") as f:
            self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop_IO"] = f
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop_datetime"] = datetime.now()
        self.exp.sync()
        del self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop_to_del"]

        assert self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop"].fetch() == "some text"
        assert self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop_number"].fetch() == 42
        assert self.exp[PROPERTIES_ATTRIBUTE_SPACE]["nested/prop"].fetch() == 42
        prop_to_del_absent = False
        try:
            self.exp[PROPERTIES_ATTRIBUTE_SPACE]["prop_to_del"].fetch()
        except AttributeError:
            prop_to_del_absent = True
        assert prop_to_del_absent

    def log_std(self):
        print("stdout text1")
        print("stdout text2")
        print("stderr text1", file=sys.stderr)
        print("stderr text2", file=sys.stderr)

    def log_series(self):
        # floats
        self.exp[LOG_ATTRIBUTE_SPACE]["m1"].log(1)
        self.exp[LOG_ATTRIBUTE_SPACE]["m1"].log(2)
        self.exp[LOG_ATTRIBUTE_SPACE]["m1"].log(3)
        self.exp[LOG_ATTRIBUTE_SPACE]["m1"].log(2)
        self.exp[LOG_ATTRIBUTE_SPACE]["nested"]["m1"].log(1)

        # texts
        self.exp[LOG_ATTRIBUTE_SPACE]["m2"].log("a")
        self.exp[LOG_ATTRIBUTE_SPACE]["m2"].log("b")
        self.exp[LOG_ATTRIBUTE_SPACE]["m2"].log("c")

        # images
        im_frame = Image.open(self.img_path)
        g_img = File.as_image(im_frame)
        self.exp[LOG_ATTRIBUTE_SPACE]["g_img"].log(g_img)

    def handle_files_and_images(self):
        # image
        im_frame = Image.open(self.img_path)
        g_img = File.as_image(im_frame)
        self.exp[ARTIFACT_ATTRIBUTE_SPACE]["assigned image"] = g_img
        self.exp.wait()
        with self.with_check_if_file_appears("assigned image.png"):
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]["assigned image"].download()
        with self.with_check_if_file_appears("custom_dest.png"):
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]["assigned image"].download("custom_dest.png")

        self.exp[ARTIFACT_ATTRIBUTE_SPACE]["logged image"].log(g_img)

        with open(self.img_path, mode="r") as f:
            # self.exp[ARTIFACT_ATTRIBUTE_SPACE]['assigned image stream'] = f
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]["logged image stream"].log(f)

        # artifact
        text_file = neptune.types.File(self.text_file_path)
        self.exp[ARTIFACT_ATTRIBUTE_SPACE]["assigned file"] = text_file
        # self.exp[ARTIFACT_ATTRIBUTE_SPACE]['logged file'].log(text_file)  # wrong type
        with open(self.text_file_path, mode="r") as f:
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]["assigned file stream"] = f
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]["logged file stream"].log(f)

    def handle_directories(self):
        pass

    def finalize(self):
        return


if __name__ == "__main__":
    NewClientFeatures().run()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import os
import unittest

from mock import patch

from neptune.new import ANONYMOUS, Run, get_last_run, init_run
from neptune.new.attributes.atoms import String
from neptune.new.envs import API_TOKEN_ENV_NAME, PROJECT_ENV_NAME
from neptune.new.exceptions import MissingFieldException, NeptuneUninitializedException
from neptune.new.internal.backends.api_model import (
    Attribute,
    AttributeType,
    IntAttribute,
)
from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock
from neptune.utils import IS_WINDOWS
from tests.neptune.new.client.abstract_experiment_test_mixin import (
    AbstractExperimentTestMixin,
)
from tests.neptune.new.utils.api_experiments_factory import api_run

AN_API_RUN = api_run()


@patch("neptune.new.internal.backends.factory.HostedNeptuneBackend", NeptuneBackendMock)
class TestClientRun(AbstractExperimentTestMixin, unittest.TestCase):
    @staticmethod
    def call_init(**kwargs):
        return init_run(**kwargs)

    @classmethod
    def setUpClass(cls) -> None:
        os.environ[PROJECT_ENV_NAME] = "organization/project"
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_metadata_container",
        new=lambda _, container_id, expected_container_type: AN_API_RUN,
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
        new=lambda _, _uuid, _type: [Attribute("some/variable", AttributeType.INT)],
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_int_attribute",
        new=lambda _, _uuid, _type, _path: IntAttribute(42),
    )
    def test_read_only_mode(self):
        exp = init_run(mode="read-only", run="whatever")

        with self.assertLogs() as caplog:
            exp["some/variable"] = 13
            exp["some/other_variable"] = 11
            self.assertEqual(
                caplog.output,
                [
                    "WARNING:neptune.new.internal.operation_processors.read_only_operation_processor:"
                    "Client in read-only mode, nothing will be saved to server."
                ],
            )

        self.assertEqual(42, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_metadata_container",
        new=lambda _, container_id, expected_container_type: AN_API_RUN,
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
        new=lambda _, _uuid, _type: [Attribute("test", AttributeType.STRING)],
    )
    def test_resume(self):
        with init_run(flush_period=0.5, run="whatever") as exp:
            self.assertEqual(exp._id, AN_API_RUN.id)
            self.assertIsInstance(exp.get_structure()["test"], String)

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch("neptune.new.internal.init.run.os.path.isfile", new=lambda file: "." in file)
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    @patch(
        "neptune.new.internal.utils.os.path.abspath",
        new=lambda path: os.path.normpath("/home/user/main_dir/" + path),
    )
    @patch("neptune.new.internal.utils.os.getcwd", new=lambda: "/home/user/main_dir")
    @unittest.skipIf(IS_WINDOWS, "Linux/Mac test")
    def test_entrypoint(self):
        exp = init_run(mode="debug")
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init_run(mode="debug", source_files=[])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init_run(mode="debug", source_files=["../*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main_dir/main.py")

        exp = init_run(mode="debug", source_files=["internal/*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init_run(mode="debug", source_files=["../other_dir/*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "../main_dir/main.py")

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch("neptune.new.internal.utils.source_code.is_ipython", new=lambda: True)
    def test_entrypoint_in_interactive_python(self):
        exp = init_run(mode="debug")
        with self.assertRaises(MissingFieldException):
            exp["source_code/entrypoint"].fetch()

        exp = init_run(mode="debug", source_files=[])
        with self.assertRaises(MissingFieldException):
            exp["source_code/entrypoint"].fetch()

        exp = init_run(mode="debug", source_files=["../*"])
        with self.assertRaises(MissingFieldException):
            exp["source_code/entrypoint"].fetch()

        exp = init_run(mode="debug", source_files=["internal/*"])
        with self.assertRaises(MissingFieldException):
            exp["source_code/entrypoint"].fetch()

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch("neptune.new.internal.utils.source_code.get_common_root", new=lambda _: None)
    @patch("neptune.new.internal.init.run.os.path.isfile", new=lambda file: "." in file)
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    @patch(
        "neptune.new.internal.utils.os.path.abspath",
        new=lambda path: os.path.normpath("/home/user/main_dir/" + path),
    )
    def test_entrypoint_without_common_root(self):
        exp = init_run(mode="debug", source_files=["../*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "/home/user/main_dir/main.py")

        exp = init_run(mode="debug", source_files=["internal/*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "/home/user/main_dir/main.py")

    def test_last_exp_is_raising_exception_when_non_initialized(self):
        # given uninitialized run
        Run.last_run = None

        # expect: raises NeptuneUninitializedException
        with self.assertRaises(NeptuneUninitializedException):
            get_last_run()

    def test_last_exp_is_the_latest_initialized(self):
        # given two initialized runs
        with init_run() as exp1, init_run() as exp2:
            # expect: `neptune.latest_run` to be the latest initialized one
            self.assertIsNot(exp1, get_last_run())
            self.assertIs(exp2, get_last_run())



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import pathlib
import shutil
import tempfile
import unittest
from pathlib import Path

from neptune.new.exceptions import (
    NeptuneLocalStorageAccessException,
    NeptuneUnsupportedArtifactFunctionalityException,
)
from neptune.new.internal.artifacts.drivers.local import LocalArtifactDriver
from neptune.new.internal.artifacts.types import (
    ArtifactDriversMap,
    ArtifactFileData,
    ArtifactFileType,
)
from tests.neptune.new.internal.artifacts.utils import md5


class TestLocalArtifactDrivers(unittest.TestCase):
    test_dir = None

    def setUp(self):
        self.test_sources_dir = Path(str(tempfile.mktemp()))
        self.test_dir = Path(str(tempfile.mktemp()))
        test_source_data = Path(__file__).parents[5] / "data" / "local_artifact_drivers_data"
        test_data = self.test_dir / "data"

        # copy source data to temp dir (to prevent e.g. inter-fs symlinks)
        shutil.copytree(test_source_data, self.test_sources_dir)

        # create files to track
        shutil.copytree(self.test_sources_dir / "files_to_track", test_data)

        # symbolic and hard link files
        # `link_to` is new in python 3.8
        # (test_source_data / 'file_to_link.txt').link_to(test_data / 'hardlinked_file.txt')
        os.link(
            src=str(self.test_sources_dir / "file_to_link.txt"),
            dst=str(test_data / "hardlinked_file.txt"),
        )
        (test_data / "symlinked_file.txt").symlink_to(self.test_sources_dir / "file_to_link.txt")

        # symlink dir - content of this file won't be discovered
        (test_data / "symlinked_dir").symlink_to(
            self.test_sources_dir / "dir_to_link", target_is_directory=True
        )

    def tearDown(self) -> None:
        # clean tmp directories
        shutil.rmtree(self.test_dir, ignore_errors=True)
        shutil.rmtree(self.test_sources_dir, ignore_errors=True)

    def test_match_by_path(self):
        self.assertEqual(ArtifactDriversMap.match_path("file:///path/to/"), LocalArtifactDriver)
        self.assertEqual(ArtifactDriversMap.match_path("/path/to/"), LocalArtifactDriver)

    def test_match_by_type(self):
        self.assertEqual(ArtifactDriversMap.match_type("Local"), LocalArtifactDriver)

    def test_file_download(self):
        path = (self.test_dir / "data/file1.txt").as_posix()
        artifact_file = ArtifactFileData(
            file_path="data/file1.txt",
            file_hash="??",
            type="??",
            metadata={"file_path": f"file://{path}"},
        )

        with tempfile.TemporaryDirectory() as temporary:
            downloaded_file = Path(temporary) / "downloaded_file.ext"

            LocalArtifactDriver.download_file(
                destination=downloaded_file, file_definition=artifact_file
            )

            self.assertTrue(Path(downloaded_file).is_symlink())
            self.assertEqual("ad62f265e5b1a2dc51f531e44e748aa0", md5(downloaded_file))

    def test_non_existing_file_download(self):
        path = "/wrong/path"
        artifact_file = ArtifactFileData(
            file_path=path, file_hash="??", type="??", metadata={"file_path": path}
        )

        with self.assertRaises(
            NeptuneLocalStorageAccessException
        ), tempfile.TemporaryDirectory() as temporary:
            local_destination = Path(temporary)
            LocalArtifactDriver.download_file(
                destination=local_destination, file_definition=artifact_file
            )

    def test_single_retrieval(self):
        files = LocalArtifactDriver.get_tracked_files(str(self.test_dir / "data/file1.txt"))

        self.assertEqual(1, len(files))
        self.assertIsInstance(files[0], ArtifactFileData)
        self.assertEqual(ArtifactFileType.LOCAL.value, files[0].type)
        self.assertEqual("d2c24d65e1d3870f4cf2dbbd8c994b4977a7c384", files[0].file_hash)
        self.assertEqual("file1.txt", files[0].file_path)
        self.assertEqual(21, files[0].size)
        self.assertEqual({"file_path", "last_modified"}, files[0].metadata.keys())
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/file1.txt').as_posix()}",
            files[0].metadata["file_path"],
        )
        self.assertIsInstance(files[0].metadata["last_modified"], str)

    def test_multiple_retrieval(self):
        files = LocalArtifactDriver.get_tracked_files(str(self.test_dir / "data"))
        files = sorted(files, key=lambda file: file.file_path)

        self.assertEqual(4, len(files))

        self.assertEqual("file1.txt", files[0].file_path)
        self.assertEqual("d2c24d65e1d3870f4cf2dbbd8c994b4977a7c384", files[0].file_hash)
        self.assertEqual(21, files[0].size)
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/file1.txt').as_posix()}",
            files[0].metadata["file_path"],
        )

        self.assertEqual("hardlinked_file.txt", files[1].file_path)
        self.assertEqual("da3e6ddfa171e1ab5564609caa1dbbea9871886e", files[1].file_hash)
        self.assertEqual(44, files[1].size)
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/hardlinked_file.txt').as_posix()}",
            files[1].metadata["file_path"],
        )

        self.assertEqual("sub_dir/file_in_subdir.txt", files[2].file_path)
        self.assertEqual("98181b1a4c880a462fcfa96b92c84b8e945ac335", files[2].file_hash)
        self.assertEqual(24, files[2].size)
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/sub_dir/file_in_subdir.txt').as_posix()}",
            files[2].metadata["file_path"],
        )

        self.assertEqual("symlinked_file.txt", files[3].file_path)
        self.assertEqual("da3e6ddfa171e1ab5564609caa1dbbea9871886e", files[3].file_hash)
        self.assertEqual(44, files[3].size)
        self.assertEqual(
            f"file://{(self.test_sources_dir.resolve() / 'file_to_link.txt').as_posix()}",
            files[3].metadata["file_path"],
        )

    def test_multiple_retrieval_prefix(self):
        files = LocalArtifactDriver.get_tracked_files(
            (self.test_dir / "data").as_posix(), "my/custom_path"
        )
        files = sorted(files, key=lambda file: file.file_path)

        self.assertEqual(4, len(files))

        self.assertEqual("my/custom_path/file1.txt", files[0].file_path)
        self.assertEqual("d2c24d65e1d3870f4cf2dbbd8c994b4977a7c384", files[0].file_hash)
        self.assertEqual(21, files[0].size)
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/file1.txt').as_posix()}",
            files[0].metadata["file_path"],
        )

        self.assertEqual("my/custom_path/hardlinked_file.txt", files[1].file_path)
        self.assertEqual("da3e6ddfa171e1ab5564609caa1dbbea9871886e", files[1].file_hash)
        self.assertEqual(44, files[1].size)
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/hardlinked_file.txt').as_posix()}",
            files[1].metadata["file_path"],
        )

        self.assertEqual("my/custom_path/sub_dir/file_in_subdir.txt", files[2].file_path)
        self.assertEqual("98181b1a4c880a462fcfa96b92c84b8e945ac335", files[2].file_hash)
        self.assertEqual(24, files[2].size)
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/sub_dir/file_in_subdir.txt').as_posix()}",
            files[2].metadata["file_path"],
        )

        self.assertEqual("my/custom_path/symlinked_file.txt", files[3].file_path)
        self.assertEqual("da3e6ddfa171e1ab5564609caa1dbbea9871886e", files[3].file_hash)
        self.assertEqual(44, files[3].size)
        self.assertEqual(
            f"file://{(self.test_sources_dir.resolve() / 'file_to_link.txt').as_posix()}",
            files[3].metadata["file_path"],
        )

    def test_expand_user(self):
        os.environ["HOME"] = str(self.test_dir.resolve())

        with open(pathlib.Path("~/tmp_test_expand_user").expanduser(), "w") as f:
            f.write("File to test ~ resolution")

        files = LocalArtifactDriver.get_tracked_files("~/tmp_test_expand_user")

        self.assertEqual(1, len(files))
        file = files[0]

        self.assertEqual("tmp_test_expand_user", file.file_path)
        self.assertEqual("eb596bf2f5fd0461d3d0b432f805b3984786c721", file.file_hash)
        self.assertEqual(25, file.size)

    def test_wildcards_not_supported(self):
        with self.assertRaises(NeptuneUnsupportedArtifactFunctionalityException):
            LocalArtifactDriver.get_tracked_files(str(self.test_dir / "data/*.txt"))



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from e2e_tests.base import BaseE2ETest
from neptune.new.exceptions import NeptuneCannotChangeStageManually
from neptune.new.metadata_containers import ModelVersion


class TestStageTransitions(BaseE2ETest):
    @pytest.mark.parametrize("container", ["model_version"], indirect=True)
    def test_transitions(self, container: ModelVersion):
        container["a"] = 14

        assert container["sys/stage"].fetch() == "none"

        container.change_stage("staging")
        container.sync()

        assert container["sys/stage"].fetch() == "staging"

        container.change_stage("production")
        container.sync()

        assert container["sys/stage"].fetch() == "production"

        container.change_stage("none")
        container.sync()

        assert container["sys/stage"].fetch() == "none"

    @pytest.mark.parametrize("container", ["model_version"], indirect=True)
    def test_fail_on_unknown_stage_value(self, container: ModelVersion):
        with pytest.raises(ValueError):
            container.change_stage("unknown")
            container.sync()

    @pytest.mark.parametrize("container", ["model_version"], indirect=True)
    def test_fail_on_manual(self, container: ModelVersion):
        with pytest.raises(NeptuneCannotChangeStageManually):
            container["sys/stage"] = "staging"
            container.sync()
